class CognitoJwtException(Exception):
    """
    Base except which all flask_jwt_extended errors extend
    """
    pass


class JWTDecodeError(CognitoJwtException):
    """
    An error decoding a JWT
    """
    pass


class InvalidHeaderError(CognitoJwtException):
    """
    An error getting header information from a request
    """
    pass


class NoAuthorizationError(CognitoJwtException):
    """
    An error raised when no authorization token was found in a
    protected endpoint
    """
    pass


class CSRFError(CognitoJwtException):
    """
    An error with CSRF protection
    """
    pass


class WrongTokenError(CognitoJwtException):
    """
    Error raised when attempting to use a refresh token to access an endpoint
    or vice versa
    """
    pass


class RevokedTokenError(CognitoJwtException):
    """
    Error raised when a revoked token attempt to access a protected endpoint
    """
    pass


class UserLoadError(CognitoJwtException):
    """
    Error raised when a user_loader callback function returns None, indicating
    that it cannot or will not load a user for the given identity.
    """
    pass


class AuthorizationExchangeError(CognitoJwtException):
    """
    An error raised when tokens could not be obtained
    in exchange for the authorization code
    """
    pass


class UserInfoError(CognitoJwtException):
    """
    An error raised when tokens could not be obtained
    in exchange for the authorization code
    """
    pass


class FlaskCognitoException(CognitoJwtException):
    """
    An error raised when tokens could not be obtained
    in exchange for the authorization code
    """
    pass



from .cognito_manager import CognitoManager
from .utils import (
    current_user, decode_token, get_csrf_token, get_current_user, get_raw_jwt,
    get_jti, get_jwt_identity, set_access_cookies,
    set_refresh_cookies, unset_access_cookies, unset_jwt_cookies,
    unset_refresh_cookies, get_unverified_jwt_headers, get_raw_jwt_header,
    get_user_info, get_encoded_access_token, get_encoded_id_token,
    get_encoded_refresh_token, get_unverified_jwt_claims
)
from .decorators import (
    jwt_optional, jwt_refresh_token_required, jwt_required, verify_jwt_in_request,
    verify_jwt_in_request_optional, verify_jwt_refresh_token_in_request,
    callback_handler, login_handler
)

__version__ = '0.2'



"""
These are the default methods implementations that are used in this extension.
All of these can be updated on an app by app basis using the CognitoManager
loader decorators. For further information, check out the following links:

"""
from flask import jsonify
import requests
import json

from flask_cognito_extended.exceptions import FlaskCognitoException
from flask_cognito_extended.config import cognito_config


def default_expired_token_callback(expired_token):
    """
    By default, if an expired token attempts to access a protected endpoint,
    we return a generic error message with a 401 status
    """
    return jsonify({cognito_config.error_msg_key: 'Token has expired'}), 401


def default_invalid_token_callback(error_string):
    """
    By default, if an invalid token attempts to access a protected endpoint, we
    return the error string for why it is not valid with a 422 status code

    :param error_string: String indicating why the token is invalid
    """
    return jsonify({cognito_config.error_msg_key: error_string}), 422


def default_unauthorized_callback(error_string):
    """
    By default, if a protected endpoint is accessed without a JWT,
    we return the error string indicating why this is unauthorized,
    with a 401 status code.

    :param error_string: String indicating why this request is unauthorized
    """
    return jsonify({cognito_config.error_msg_key: error_string}), 401


def default_revoked_token_callback():
    """
    By default, if a revoked token is used to access a protected endpoint, we
    return a general error message with a 401 status code
    """
    return jsonify({cognito_config.error_msg_key: 'Token has been'
                                                  ' revoked'}), 401


def default_user_loader_error_callback(identity):
    """
    By default, if a user_loader callback is defined and the callback
    function returns None, we return a general error message with a 401
    status code
    """
    result = {cognito_config.error_msg_key: "Error loading the"
                                            " user {}".format(identity)}
    return jsonify(result), 401


def default_decode_key_callback(public_uri):
    """
    The default implementation returns the public key provided by Cognito
    at the endpoint:
    https://cognito-idp.{region}.amazonaws.com/{userPoolId}/.well-known/jwks.json
    """
    try:
        response = requests.get(public_uri)
        key = json.loads(response.text)
    except requests.exceptions.RequestException as e:
        raise FlaskCognitoException(str(e)) from e
    return key


def default_authorization_failed_callback(error_string):
    """
    By default, if a authorization code exchange fails, we will return
    the message passed by cognito with a 401 status code
    """
    return jsonify({cognito_config.error_msg_key: error_string}), 401


def default_user_endpoint_error_callback(error_string):
    """
    By default, if a user info endpoint fails, we will return
    the message passed by cognito with a 401 status code
    """
    return jsonify({cognito_config.error_msg_key: error_string}), 401


def default_general_error_callback(error_string):
    """
    By default, if a user info endpoint fails, we will return
    the message passed by cognito with a 401 status code
    """
    return jsonify({cognito_config.error_msg_key: error_string}), 401



import datetime
from hashlib import md5
from six import raise_from
from flask import current_app
try:
    from collections.abc import Sequence, Set
except ImportError:  # pragma: no cover
    from collections import Sequence, Set


class _Config(object):
    """
    Helper object for accessing and verifying options in this extension. This
    is meant for internal use of the application.

    Default values for the configuration options are set in the cognito_manager
    object. All of these values are read only. This is simply a loose wrapper
    with some helper functionality for flasks `app.config`.
    """

    @property
    def scope(self):
        scope = current_app.config['COGNITO_SCOPE']
        if not scope:
            raise RuntimeError('COGNITO_SCOPE must be specified to '
                               'create the auth url. ')
        return scope

    @property
    def state(self):
        return md5("{}:{}".format(self.client_id, self.user_pool_id).
                   encode("utf-8")).hexdigest()

    @property
    def issuer(self):
        return "https://cognito-idp.{region}.amazonaws.com/{userPoolId}".format(
                            region=self.region, userPoolId=self.user_pool_id)

    @property
    def region(self):
        region = current_app.config['COGNITO_REGION']
        if not region:
            raise RuntimeError('COGNITO_REGION must be specified.')
        return region

    @property
    def public_key_uri(self):
        return ("https://cognito-idp.{region}.amazonaws.com"
                "/{userpoolid}/.well-known/jwks.json").format(
                region=self.region, userpoolid=self.user_pool_id)

    @property
    def login_uri(self):
        return ("{domain}/login?client_id={client_id}"
                "&response_type=code&scope={scope}"
                "&redirect_uri={redirect_uri}&state={state}").format(
                domain=self.domain, client_id=self.client_id,
                scope=self.scope, redirect_uri=self.redirect_uri,
                state=self.state)

    @property
    def user_pool_id(self):
        pool_id = current_app.config['COGNITO_USER_POOL_ID']
        if not pool_id:
            raise RuntimeError('COGNITO_USER_POOL_ID must be specified to '
                               'locate the auth url. ')
        return pool_id

    @property
    def client_id(self):
        client_id = current_app.config['COGNITO_CLIENT_ID']
        if not client_id:
            raise RuntimeError('COGNITO_CLIENT_ID must be set to '
                               'validate the audience claim. ')
        return client_id

    @property
    def domain(self):
        domain = current_app.config['COGNITO_DOMAIN']
        if not domain:
            raise RuntimeError('COGNITO_DOMAIN must be set to '
                               'redirect to create endpoint url. ')
        if not domain.startswith("https://"):
            domain = "https://{}".format(domain)
        return domain

    @property
    def redirect_uri(self):
        uri = current_app.config['COGNITO_REDIRECT_URI']
        if not uri:
            raise RuntimeError('COGNITO_REDIRECT_URI must be set to '
                               'obtain callback url. ')
        return uri

    @property
    def exempt_methods(self):
        methods = current_app.config['EXEMPT_METHODS']
        return methods if methods else {"OPTIONS"}

    @property
    def identity_claim_key(self):
        key = current_app.config['JWT_IDENTITY_CLAIM']
        return key if key else 'sub'

    @property
    def token_location(self):
        locations = current_app.config['JWT_TOKEN_LOCATION']
        if isinstance(locations, str):
            locations = (locations,)
        elif not isinstance(locations, (Sequence, Set)):
            raise RuntimeError('JWT_TOKEN_LOCATION must be a'
                               ' sequence or a set')
        elif not locations:
            raise RuntimeError('JWT_TOKEN_LOCATION must contain at least one '
                               'of "headers", "cookies", "query_string" or "json"')
        for location in locations:
            if location not in ('headers', 'cookies', 'json', 'query_string'):
                raise RuntimeError('JWT_TOKEN_LOCATION can only contain '
                                   '"headers", "cookies", "query_string" or "json"')
        return locations

    @property
    def jwt_in_cookies(self):
        return 'cookies' in self.token_location

    @property
    def jwt_in_headers(self):
        return 'headers' in self.token_location

    @property
    def jwt_in_query_string(self):
        return 'query_string' in self.token_location

    @property
    def jwt_in_json(self):
        return 'json' in self.token_location

    @property
    def error_msg_key(self):
        return current_app.config['JWT_ERROR_MESSAGE_KEY']

    @property
    def header_name(self):
        name = current_app.config['JWT_HEADER_NAME']
        if not name:
            raise RuntimeError("JWT_ACCESS_HEADER_NAME cannot be empty")
        return name

    @property
    def header_type(self):
        return current_app.config['JWT_HEADER_TYPE']

    @property
    def query_string_name(self):
        return current_app.config['JWT_QUERY_STRING_NAME']

    @property
    def access_cookie_name(self):
        return current_app.config['JWT_ACCESS_COOKIE_NAME']

    @property
    def refresh_cookie_name(self):
        return current_app.config['JWT_REFRESH_COOKIE_NAME']

    @property
    def access_cookie_path(self):
        return current_app.config['JWT_ACCESS_COOKIE_PATH']

    @property
    def refresh_cookie_path(self):
        return current_app.config['JWT_REFRESH_COOKIE_PATH']

    @property
    def cookie_secure(self):
        return current_app.config['JWT_COOKIE_SECURE']

    @property
    def cookie_domain(self):
        return current_app.config['JWT_COOKIE_DOMAIN']

    @property
    def session_cookie(self):
        return current_app.config['JWT_SESSION_COOKIE']

    @property
    def cookie_samesite(self):
        return current_app.config['JWT_COOKIE_SAMESITE']

    @property
    def json_key(self):
        return current_app.config['JWT_JSON_KEY']

    @property
    def refresh_json_key(self):
        return current_app.config['JWT_REFRESH_JSON_KEY']

    @property
    def csrf_protect(self):
        return self.jwt_in_cookies and current_app.config['JWT_COOKIE_CSRF_PROTECT']

    @property
    def csrf_request_methods(self):
        return current_app.config['JWT_CSRF_METHODS']

    @property
    def csrf_in_cookies(self):
        return current_app.config['JWT_CSRF_IN_COOKIES']

    @property
    def access_csrf_cookie_name(self):
        return current_app.config['JWT_ACCESS_CSRF_COOKIE_NAME']

    @property
    def refresh_csrf_cookie_name(self):
        return current_app.config['JWT_REFRESH_CSRF_COOKIE_NAME']

    @property
    def access_csrf_header_name(self):
        return current_app.config['JWT_ACCESS_CSRF_HEADER_NAME']

    @property
    def access_csrf_cookie_path(self):
        return current_app.config['JWT_ACCESS_CSRF_COOKIE_PATH']

    @property
    def refresh_csrf_cookie_path(self):
        return current_app.config['JWT_REFRESH_CSRF_COOKIE_PATH']

    @property
    def refresh_csrf_header_name(self):
        return current_app.config['JWT_REFRESH_CSRF_HEADER_NAME']

    @property
    def access_csrf_field_name(self):
        return current_app.config['JWT_ACCESS_CSRF_FIELD_NAME']

    @property
    def refresh_csrf_field_name(self):
        return current_app.config['JWT_REFRESH_CSRF_FIELD_NAME']

    @property
    def csrf_check_form(self):
        return current_app.config['JWT_CSRF_CHECK_FORM']

    @property
    def access_expires(self):
        delta = current_app.config['JWT_ACCESS_TOKEN_EXPIRES']
        if type(delta) is int:
            delta = datetime.timedelta(seconds=delta)
        if delta is not False:
            try:
                delta + datetime.datetime.now()
            except TypeError as e:
                err = (
                    "must be able to add JWT_ACCESS_TOKEN_EXPIRES to datetime.datetime"
                )
                raise_from(RuntimeError(err), e)
        return delta

    @property
    def refresh_expires(self):
        delta = current_app.config['JWT_REFRESH_TOKEN_EXPIRES']
        if type(delta) is int:
            delta = datetime.timedelta(seconds=delta)
        if delta is not False:
            try:
                delta + datetime.datetime.now()
            except TypeError as e:
                err = (
                    "must be able to add JWT_REFRESH_TOKEN_EXPIRES to datetime.datetime"
                )
                raise_from(RuntimeError(err), e)
        return delta

    @property
    def blacklist_enabled(self):
        return current_app.config['JWT_BLACKLIST_ENABLED']

    @property
    def blacklist_checks(self):
        check_type = current_app.config['JWT_BLACKLIST_TOKEN_CHECKS']
        if isinstance(check_type, str):
            check_type = (check_type,)
        elif not isinstance(check_type, (Sequence, Set)):
            raise RuntimeError('JWT_BLACKLIST_TOKEN_CHECKS must be a sequence or a set')
        for item in check_type:
            if item not in ('access', 'refresh'):
                err = 'JWT_BLACKLIST_TOKEN_CHECKS must be "access" or "refresh"'
                raise RuntimeError(err)
        return check_type

    @property
    def blacklist_access_tokens(self):
        return 'access' in self.blacklist_checks

    @property
    def blacklist_refresh_tokens(self):
        return 'refresh' in self.blacklist_checks

    @property
    def cookie_max_age(self):
        # Returns the appropiate value for max_age for flask set_cookies. If
        # session cookie is true, return None, otherwise return a number of
        # seconds 1 year in the future
        return None if self.session_cookie else 31540000  # 1 year

    @property
    def audience(self):
        return self.client_id


cognito_config = _Config()



"""
Flask-Cognito-Extended
-------------

Flask-Cognito-Extended is a Flask implementation of Amazon Cognito.
"""
import io
import re
from setuptools import setup

with io.open('flask_cognito_extended/__init__.py', encoding='utf-8') as f:
    version = re.search(r"__version__ = '(.+)'", f.read()).group(1)

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='Flask-Cognito-Extended',
    version=version,
    url='https://github.com/deejungx/flask-cognito-extended',
    license='MIT',
    author='Dipesh Jung Pandey',
    author_email='dipzz1394@gmail.com',
    description='Extended Cognito integration with Flask',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=['flask', 'amazon cognito', 'json web token'],
    packages=['flask_cognito_extended'],
    zip_safe=False,
    platforms='any',
    install_requires=[
        'Flask>=1.0,<2.0',
        'Werkzeug>=0.14',
        'python-jose[cryptography]>=3.1.0',
        'PyJWT>=1.7,<2.0',
        'six',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Version Control :: Git',
        'Topic :: System :: Systems Administration :: Authentication/Directory'
    ]
)


from werkzeug.local import LocalProxy
from jose import jwt
import json
import requests
from datetime import datetime, timedelta
import email.utils as eut
from flask import current_app, redirect
from flask_cognito_extended.tokens import decode_jwt
from flask_cognito_extended.config import cognito_config
from flask_cognito_extended.exceptions import (
    WrongTokenError, RevokedTokenError, AuthorizationExchangeError,
    UserInfoError)
try:
    from flask import _app_ctx_stack as ctx_stack
except ImportError:  # pragma: no cover
    from flask import _request_ctx_stack as ctx_stack

# Proxy to access the current user
current_user = LocalProxy(lambda: get_current_user())


def get_user_info(access_token):
    """
    Calls the userInfo endpoint on Cognito.
    Returns a dictionary containing user info.

    :param access_token: encoded access token
    :return: Dictionary containing user info
    """
    user_url = "{}/oauth2/userInfo".format(cognito_config.domain)
    headers = {"Authorization": "Bearer {}".format(access_token)}
    try:
        response = requests.post(user_url, headers=headers)
        response_json = response.json()
    except requests.exceptions.RequestException as e:
        raise UserInfoError(str(e)) from e
    return response_json


def exchange_code_for_token(code):
    """
    Exchanges authorization code for tokens using the Cogntio TOKEN endpoint.
    This includes the encoded id, access and refresh tokens.

    :param code: The authorization code obtained from user login/sign-up
                 confimation.
    :return: Dictionary containing encoded access_token, refresh_token
             and id_token
    """
    url = '{domain}/oauth2/token'.format(domain=cognito_config.domain)
    headers = {'Content-type': 'application/x-www-form-urlencoded'}
    data = {'grant_type': 'authorization_code',
            'client_id': cognito_config.client_id,
            'redirect_uri': cognito_config.redirect_uri,
            'code': code
            }
    try:
        response = requests.post(url=url, headers=headers, data=data)
        tokens = json.loads(response.text)
    except requests.exceptions.HTTPError as e:
        raise AuthorizationExchangeError(str(e)) from e
    # check token errors
    # check token expiry
    published_time = datetime(*eut.parsedate(response.headers['Date'])[:6])
    expiry = tokens.pop('expires_in')
    expiry_time = published_time + timedelta(int(expiry))
    if datetime.now() > expiry_time:
        raise AuthorizationExchangeError("Request is expired")
    # check token type bearer
    token_type = tokens.pop('token_type')
    if token_type != 'Bearer':
        raise AuthorizationExchangeError("Invalid token type")
    return tokens


def get_encoded_access_token():
    """
    The encoded access token obtained from exchanging the code
    is accessed from this endpoint. If no token string is found,
    an empty string is returned instead.
    """
    return getattr(ctx_stack.top, 'encoded_access_token', "")


def get_encoded_refresh_token():
    """
    The encoded refresh token obtained from exchanging the code
    is accessed from this endpoint. If no token string is found,
    an empty string is returned instead.
    """
    return getattr(ctx_stack.top, 'encoded_refresh_token', "")


def get_encoded_id_token():
    """
    The encoded id token obtained from exchanging the code
    is accessed from this endpoint. If no token string is found,
    an empty string is returned instead.
    """
    return getattr(ctx_stack.top, 'encoded_id_token', "")


def get_raw_jwt():
    """
    In a protected endpoint, this will return the python dictionary which has
    all of the claims of the JWT that is accessing the endpoint. If no
    JWT is currently present, an empty dict is returned instead.
    """
    return getattr(ctx_stack.top, 'jwt', {})


def get_raw_jwt_header():
    """
    In a protected endpoint, this will return the python dictionary which has
    the JWT headers values. If no
    JWT is currently present, an empty dict is returned instead.
    """
    return getattr(ctx_stack.top, 'jwt_header', {})


def get_jwt_identity():
    """
    In a protected endpoint, this will return the identity of the JWT that is
    accessing this endpoint. If no JWT is present,`None` is returned instead.
    """
    return get_raw_jwt().get(cognito_config.identity_claim_key, None)


def get_current_user():
    """
    In a protected endpoint, this will return the user object for the JWT that
    is accessing this endpoint. This is only present if the
    :meth:`~flask_jwt_extended.CognitoManager.user_loader_callback_loader` is
    being used. If the user loader callback is not being used, this will
    return `None`.
    """
    return getattr(ctx_stack.top, 'jwt_user', None)


def get_jti(encoded_token):
    """
    Returns the JTI (unique identifier) of an encoded JWT

    :param encoded_token: The encoded JWT to get the JTI from.
    """
    return decode_token(encoded_token).get('jti')


def decode_token(encoded_token, csrf_value=None, allow_expired=False):
    """
    Returns the decoded token (python dict) from an encoded JWT. This does all
    the checks to insure that the decoded token is valid before returning it.

    :param encoded_token: The encoded JWT to decode into a python dict.
    :param csrf_value: Expected CSRF double submit value (optional)
    :param allow_expired: Options to ignore exp claim validation in token
    :return: Dictionary containing contents of the JWT
    """
    cognito_manager = _get_cognito_manager()
    secret = cognito_manager._decode_key_callback(cognito_config.public_key_uri)['keys']

    return decode_jwt(
        encoded_token=encoded_token,
        secret=secret,
        identity_claim_key=cognito_config.identity_claim_key,
        csrf_value=csrf_value,
        audience=cognito_config.audience,
        issuer=cognito_config.issuer,
        allow_expired=allow_expired
    )


def _get_cognito_manager():
    try:
        return current_app.extensions['flask-cognito-extended']
    except KeyError:  # pragma: no cover
        raise RuntimeError("You must initialize a CognitoManager with this flask "
                           "application before using this method")


def has_user_loader():
    cognito_manager = _get_cognito_manager()
    return cognito_manager._user_loader_callback is not None


def user_loader(*args, **kwargs):
    cognito_manager = _get_cognito_manager()
    return cognito_manager._user_loader_callback(*args, **kwargs)


def has_token_in_blacklist_callback():
    cognito_manager = _get_cognito_manager()
    return cognito_manager._token_in_blacklist_callback is not None


def token_in_blacklist(*args, **kwargs):
    cognito_manager = _get_cognito_manager()
    return cognito_manager._token_in_blacklist_callback(*args, **kwargs)


def verify_token_type(decoded_token, expected_type):
    if decoded_token['type'] != expected_type:
        raise WrongTokenError('Only {} tokens are allowed'.format(expected_type))


def verify_token_not_blacklisted(decoded_token, request_type):
    if not cognito_config.blacklist_enabled:
        return
    if not has_token_in_blacklist_callback():
        raise RuntimeError("A token_in_blacklist_callback must be provided via "
                           "the '@token_in_blacklist_loader' if "
                           "JWT_BLACKLIST_ENABLED is True")
    if cognito_config.blacklist_access_tokens and request_type == 'access':
        if token_in_blacklist(decoded_token):
            raise RevokedTokenError('Token has been revoked')
    if cognito_config.blacklist_refresh_tokens and request_type == 'refresh':
        if token_in_blacklist(decoded_token):
            raise RevokedTokenError('Token has been revoked')


def get_csrf_token(encoded_token):
    """
    Returns the CSRF double submit token from an encoded JWT.

    :param encoded_token: The encoded JWT
    :return: The CSRF double submit token
    """
    token = decode_token(encoded_token)
    return token['csrf']


def set_access_cookies(response, encoded_access_token, max_age=None):
    """
    Takes a flask response object, and an encoded access token, and configures
    the response to set in the access token in a cookie. If `JWT_CSRF_IN_COOKIES`
    is `True` (see :ref:`Configuration Options`), this will also set the CSRF
    double submit values in a separate cookie.

    :param response: The Flask response object to set the access cookies in.
    :param encoded_access_token: The encoded access token to set in the cookies.
    :param max_age: The max age of the cookie. If this is None, it will use the
                    `JWT_SESSION_COOKIE` option (see :ref:`Configuration Options`).
                    Otherwise, it will use this as the cookies `max-age` and the
                    JWT_SESSION_COOKIE option will be ignored.  Values should be
                    the number of seconds (as an integer).
    """
    if not cognito_config.jwt_in_cookies:
        raise RuntimeWarning("set_access_cookies() called without "
                             "'JWT_TOKEN_LOCATION' configured to use cookies")

    # Set the access JWT in the cookie
    response.set_cookie(cognito_config.access_cookie_name,
                        value=encoded_access_token,
                        max_age=max_age or cognito_config.cookie_max_age,
                        secure=cognito_config.cookie_secure,
                        httponly=True,
                        domain=cognito_config.cookie_domain,
                        path=cognito_config.access_cookie_path,
                        samesite=cognito_config.cookie_samesite)

    # If enabled, set the csrf double submit access cookie
    if cognito_config.csrf_protect and cognito_config.csrf_in_cookies:
        response.set_cookie(cognito_config.access_csrf_cookie_name,
                            value=get_csrf_token(encoded_access_token),
                            max_age=max_age or cognito_config.cookie_max_age,
                            secure=cognito_config.cookie_secure,
                            httponly=False,
                            domain=cognito_config.cookie_domain,
                            path=cognito_config.access_csrf_cookie_path,
                            samesite=cognito_config.cookie_samesite)


def set_refresh_cookies(response, encoded_refresh_token, max_age=None):
    """
    Takes a flask response object, and an encoded refresh token, and configures
    the response to set in the refresh token in a cookie. If `JWT_CSRF_IN_COOKIES`
    is `True` (see :ref:`Configuration Options`), this will also set the CSRF
    double submit values in a separate cookie.

    :param response: The Flask response object to set the refresh cookies in.
    :param encoded_refresh_token: The encoded refresh token to set in the cookies.
    :param max_age: The max age of the cookie. If this is None, it will use the
                    `JWT_SESSION_COOKIE` option (see :ref:`Configuration Options`).
                    Otherwise, it will use this as the cookies `max-age` and the
                    JWT_SESSION_COOKIE option will be ignored.  Values should be
                    the number of seconds (as an integer).
    """
    if not cognito_config.jwt_in_cookies:
        raise RuntimeWarning("set_refresh_cookies() called without "
                             "'JWT_TOKEN_LOCATION' configured to use cookies")

    # Set the refresh JWT in the cookie
    response.set_cookie(cognito_config.refresh_cookie_name,
                        value=encoded_refresh_token,
                        max_age=max_age or cognito_config.cookie_max_age,
                        secure=cognito_config.cookie_secure,
                        httponly=True,
                        domain=cognito_config.cookie_domain,
                        path=cognito_config.refresh_cookie_path,
                        samesite=cognito_config.cookie_samesite)

    # If enabled, set the csrf double submit refresh cookie
    if cognito_config.csrf_protect and cognito_config.csrf_in_cookies:
        response.set_cookie(cognito_config.refresh_csrf_cookie_name,
                            value=get_csrf_token(encoded_refresh_token),
                            max_age=max_age or cognito_config.cookie_max_age,
                            secure=cognito_config.cookie_secure,
                            httponly=False,
                            domain=cognito_config.cookie_domain,
                            path=cognito_config.refresh_csrf_cookie_path,
                            samesite=cognito_config.cookie_samesite)


def unset_jwt_cookies(response):
    """
    Takes a flask response object, and configures it to unset (delete) JWTs
    stored in cookies.

    :param response: The Flask response object to delete the JWT cookies in.
    """
    unset_access_cookies(response)
    unset_refresh_cookies(response)


def unset_access_cookies(response):
    """
    takes a flask response object, and configures it to unset (delete) the
    access token from the response cookies. if `jwt_csrf_in_cookies`
    (see :ref:`configuration options`) is `true`, this will also remove the
    access csrf double submit value from the response cookies as well.

    :param response: the flask response object to delete the jwt cookies in.
    """
    if not cognito_config.jwt_in_cookies:
        raise RuntimeWarning("unset_access_cookies() called without "
                             "'JWT_TOKEN_LOCATION' configured to use cookies")

    response.set_cookie(cognito_config.access_cookie_name,
                        value='',
                        expires=0,
                        secure=cognito_config.cookie_secure,
                        httponly=True,
                        domain=cognito_config.cookie_domain,
                        path=cognito_config.access_cookie_path,
                        samesite=cognito_config.cookie_samesite)

    if cognito_config.csrf_protect and cognito_config.csrf_in_cookies:
        response.set_cookie(cognito_config.access_csrf_cookie_name,
                            value='',
                            expires=0,
                            secure=cognito_config.cookie_secure,
                            httponly=False,
                            domain=cognito_config.cookie_domain,
                            path=cognito_config.access_csrf_cookie_path,
                            samesite=cognito_config.cookie_samesite)


def unset_refresh_cookies(response):
    """
    takes a flask response object, and configures it to unset (delete) the
    refresh token from the response cookies. if `jwt_csrf_in_cookies`
    (see :ref:`configuration options`) is `true`, this will also remove the
    refresh csrf double submit value from the response cookies as well.

    :param response: the flask response object to delete the jwt cookies in.
    """
    if not cognito_config.jwt_in_cookies:
        raise RuntimeWarning("unset_refresh_cookies() called without "
                             "'JWT_TOKEN_LOCATION' configured to use cookies")

    response.set_cookie(cognito_config.refresh_cookie_name,
                        value='',
                        expires=0,
                        secure=cognito_config.cookie_secure,
                        httponly=True,
                        domain=cognito_config.cookie_domain,
                        path=cognito_config.refresh_cookie_path,
                        samesite=cognito_config.cookie_samesite)

    if cognito_config.csrf_protect and cognito_config.csrf_in_cookies:
        response.set_cookie(cognito_config.refresh_csrf_cookie_name,
                            value='',
                            expires=0,
                            secure=cognito_config.cookie_secure,
                            httponly=False,
                            domain=cognito_config.cookie_domain,
                            path=cognito_config.refresh_csrf_cookie_path,
                            samesite=cognito_config.cookie_samesite)


def get_unverified_jwt_headers(encoded_token):
    """
    Returns the Headers of an encoded JWT without verifying the actual signature of JWT.
     Note: The signature is not verified so the header parameters
     should not be fully trusted until signature verification is complete

    :param encoded_token: The encoded JWT to get the Header from.
    :return: JWT header parameters as python dict()
    """
    return jwt.get_unverified_header(encoded_token)


def get_unverified_jwt_claims(encoded_token):
    """
    Returns the Headers of an encoded JWT without verifying the actual signature of JWT.
     Note: The signature is not verified so the header parameters
     should not be fully trusted until signature verification is complete

    :param encoded_token: The encoded JWT to get the Header from.
    :return: JWT header parameters as python dict()
    """
    return jwt.get_unverified_claims(encoded_token)



import time
from jose import jwk, jwt
from jose.utils import base64url_decode
from werkzeug.security import safe_str_cmp
from flask_cognito_extended.exceptions import CSRFError, JWTDecodeError
try:
    from flask import _app_ctx_stack as ctx_stack
except ImportError:  # pragma: no cover
    from flask import _request_ctx_stack as ctx_stack


def decode_jwt(encoded_token, secret, identity_claim_key,
               csrf_value=None, audience=None, allow_expired=False,
               issuer=None):
    """
    Decodes an encoded JWT

    :param encoded_token: The encoded JWT string to decode
    :param secret: Secret key used to decode the JWT
    :param identity_claim_key: expected key that contains the identity
    :param csrf_value: Expected double submit csrf value
    :param audience: expected audience in the JWT
    :param issuer: expected issuer in the JWT
    :param allow_expired: Options to ignore exp claim validation in token
    :return: Dictionary containing contents of the JWT
    """
    token = encoded_token
    headers = jwt.get_unverified_headers(token)
    kid = headers['kid']
    # search for the kid in the downloaded public keys
    key_index = -1
    for i in range(len(secret)):
        if kid == secret[i]['kid']:
            key_index = i
            break
    if key_index == -1:
        raise JWTDecodeError("Invalid key attribute: kid")
    # construct the public key
    public_key = jwk.construct(secret[key_index])
    # get the last two sections of the token,
    # message and signature (encoded in base64)
    message, encoded_signature = str(token).rsplit('.', 1)
    # decode the signature
    decoded_signature = base64url_decode(encoded_signature.encode('utf-8'))
    # verify the signature
    if not public_key.verify(message.encode("utf8"), decoded_signature):
        raise JWTDecodeError("Signature verification failed")
    # since we passed the verification, we can now safely
    # use the unverified claims
    data = jwt.get_unverified_claims(token)
    if identity_claim_key not in data:
        raise JWTDecodeError("Missing claim: {}".format(identity_claim_key))
    if not allow_expired and time.time() > data['exp']:
        ctx_stack.top.expired_jwt = token
        raise JWTDecodeError("Token has expired")
    # check iss
    if 'iss' not in data:
        data['iss'] = None
    if data['iss'] != issuer:
        raise JWTDecodeError("Missing or invalid issuer")
    # check aud
    if 'aud' not in data:
        data['aud'] = None
    if data['aud'] not in audience:
        raise JWTDecodeError("Missing or invalid audience")
    # check csrf
    if csrf_value:
        if 'csrf' not in data:
            raise JWTDecodeError("Missing claim: csrf")
        if not safe_str_cmp(data['csrf'], csrf_value):
            raise CSRFError("CSRF double submit tokens do not match")
    return data
